import unittest
import numpy as np
import matplotlib.pyplot as plt
import subprocess as sp
from scipy.optimize import curve_fit
import os 
import pytest

plotstuff = True

def fn(x, m):
    return m*x

def load_var(filename, var):
    with open(filename, 'r') as f:
        for line in f:
            if line.startswith('variable '+var):
                try:
                    out = float(line.split()[3])
                except ValueError:
                    print("Error: Unable to convert " + var + " to a float.")
                    out = None
                break
    return out

# Fixture to run the command before tests
@pytest.fixture(scope="session", autouse=True)
def run_command_before_tests():
    command = "mpiexec -n 4 python3 run_TTCF.py"
    cwd = "../"
    sp.call(command, cwd=cwd, shell=True)

def test_files_exist():
    assert os.path.isfile("../profile_DAV.txt")
    assert os.path.isfile("../profile_TTCF.txt")

def test_files_contain_data():
    #Load output profiles
    DAV = np.loadtxt("../profile_DAV.txt")
    TTCF = np.loadtxt("../profile_TTCF.txt")

    assert TTCF.shape[0] > 0
    assert DAV.shape[0] > 0

def test_analytical_soln():
    #Load output profiles
    DAV = np.loadtxt("../profile_DAV.txt")
    TTCF = np.loadtxt("../profile_TTCF.txt")

    #Analytical solution
    filename = "../system_setup.in"
    srate = load_var(filename, "srate")
    rho = load_var(filename, "rho")
    Npart = load_var(filename, "Npart")
    L = (Npart/rho)**(1.0/3.0)
    x = np.linspace(0., L, TTCF.shape[1])
    vx = srate*x

    #Compare analytical to TTCF solution numerically
    paramsTTCF, covarianceTTCF = curve_fit(fn, x[1:-1], np.mean(TTCF[:,1:-1],0), p0=[0.])
    errorTTCF = abs(paramsTTCF[0]-srate)/srate

    paramsDAV, covarianceDAV = curve_fit(fn, x[1:-1], np.mean(DAV[:,1:-1],0), p0=[0.])
    errorDAV = abs(paramsDAV[0]-srate)/srate
    print("Fitted TTCF strainrate = ", paramsTTCF[0], 
          "Fitted DAV strainrate = ", paramsDAV[0],
          " enforced strainrate = ", srate, 
          " errorTTCF = ", errorTTCF, " errorDAV = ", errorDAV,
          "covarianceTTCF=", covarianceTTCF, "covarianceDAV=", covarianceDAV, )

    #Test error less than 20%
    assert errorTTCF < 0.2

    #If strain rate is low, error should be less
    # and covariance in fit
    if srate < 1e-5:
        assert errorTTCF < errorDAV
        assert covarianceTTCF[0] < covarianceDAV[0]




import unittest
import numpy as np

from TTCF import utils

class TestUpdateFunctions(unittest.TestCase):
    def test_update_var(self):
        partial = 5.0
        mean = 3.0
        var = 2.0
        Count = 4

        result = utils.update_var(partial, mean, var, Count)

        # Calculate the expected result manually
        newmean=((Count-1)*mean + partial)/float(Count)
        expected_result = ((Count-2)*var + (partial-mean)*(partial-newmean))/float(Count-1)
        
        self.assertEqual(result, expected_result)

    def test_update_mean(self):
        partial = 5.0
        mean = 3.0
        Count = 4

        result = utils.update_mean(partial, mean, Count)

        # Calculate the expected result manually
        expected_result = ((Count - 1) * mean + partial) / float(Count)
        
        self.assertEqual(result, expected_result)

class TestTTCFIntegration(unittest.TestCase):
    def test_sum_prev_dt(self):
        A = np.array([1.0, 2.0, 3.0, 4.0])
        t = 3
        result = utils.sum_prev_dt(A, t)
        expected_result = (A[1] + 4 * A[2] + A[3]) / 3.0
        self.assertEqual(result, expected_result)

    def test_TTCF_integration(self):
        A = np.array([0.0, 0.2, 0.4, 0.6])
        int_step = 0.2

        result = utils.TTCF_integration(A, int_step)

        # Updated expected result to match the calculated result more precisely
        expected_result = np.array([0.  , 0.04, 0.08, 0.14])
        np.testing.assert_allclose(result, expected_result, rtol=1e-5)

if __name__ == '__main__':
    unittest.main()




import unittest
import numpy as np
from TTCF import TTCF

class TestTTCF(unittest.TestCase):
    def setUp(self):
        # Create a sample TTCF object for testing
        global_variables = ["var1", "var2"]
        profile_variables = ["var3", "var4"]
        Nsteps = 100
        Nbins = 10
        Nmappings = 5

        self.ttcf_obj = TTCF.TTCF(global_variables, profile_variables, Nsteps, Nbins, Nmappings)

    def test_initialization(self):
        # Test the initialization of the TTCF object
        self.assertEqual(self.ttcf_obj.Nsteps, 100)
        self.assertEqual(self.ttcf_obj.Nbins, 10)
        self.assertEqual(self.ttcf_obj.Nmappings, 5)
        self.assertEqual(self.ttcf_obj.Count, 0)

        # Ensure that arrays are initialized with the correct shape
        self.assertEqual(self.ttcf_obj.DAV_global_mean.shape, (100, 2))
        self.assertEqual(self.ttcf_obj.DAV_profile_mean.shape, (100, 10, 4))

    def test_add_mappings(self):
        # Test the add_mappings method

        # Create sample data for testing
        data_profile = np.ones((100, 10, 4))
        data_global = np.ones((100, 2))
        omega = 2.0

        # Add mappings to the TTCF object
        self.ttcf_obj.add_mappings(data_profile, data_global, omega)

        # Add assertions based on the expected behavior of your add_mappings method
        self.assertEqual(np.sum(self.ttcf_obj.DAV_profile_partial), 100 * 10 * 4)
        self.assertEqual(np.sum(self.ttcf_obj.DAV_global_partial), 100 * 2)

if __name__ == '__main__':
    unittest.main()



from mpi4py import MPI
from lammps import lammps
from lammps import LMP_STYLE_GLOBAL, LMP_TYPE_VECTOR, LMP_TYPE_ARRAY

import numpy as np

def sum_over_MPI(A, irank, comm, root=0):
    #Put data into contiguous c arrays read to send through MPI
    sendbuf = np.ascontiguousarray(A)
    if irank == root:
        recvbuf = np.copy(np.ascontiguousarray(A))
    else:
        recvbuf = np.array(1)
    comm.Reduce([sendbuf, MPI.DOUBLE], [recvbuf, MPI.DOUBLE], op=MPI.SUM, root=root )
    #Summed arrays only exist on root process, unpack into variables
    if irank == root:
        return recvbuf
    else:
        return None

def update_var(partial, mean, var, Count):
    newmean=((Count-1)*mean + partial)/float(Count)
    return ((Count-2)*var + (partial-mean)*(partial-newmean))/float(Count-1)

def update_mean(partial, mean, Count):
    return ((Count-1)*mean + partial)/float(Count)

def  run_mother_trajectory(lmp,Nsteps,Thermo_damp):

    lmp.command("fix NVT_equilibrium all nvt temp ${T} ${T} " +  str(Thermo_damp) + " tchain 1")
    lmp.command("run " + str(Nsteps))
    lmp.command("unfix NVT_equilibrium")

    return None


def get_fix_data(lmp, fixname, variables, Nbins=None):

    nlmp = lmp.numpy
    if type(variables) is list:
        Ncols = len(variables)
    elif type(variables) is int:
        Ncols = variables
    if Nbins == None:
        out = np.empty(Ncols)
        for column in range(Ncols):
            out[column] = nlmp.extract_fix(fixname, LMP_STYLE_GLOBAL, 
                                            LMP_TYPE_VECTOR, column)
    else:
        #The extra 2 is to account for Cell index and Ncount in array types (e.g. chunks)
        Nrows = Nbins
        out = np.empty([Nbins, Ncols+2])
        for row in range(Nrows):
            for column in range(Ncols+2):
                out[row, column] = nlmp.extract_fix(fixname, LMP_STYLE_GLOBAL, 
                                                    LMP_TYPE_ARRAY, row , column)

    return out

def get_globaldata(lmp, global_variables):
    return get_fix_data(lmp, "Global_variables", global_variables)

def get_globalDict(lmp, global_variables):
    """
        Otherwise we could return a dictonary 
        of named output variables as specified by the user
    """
    nlmp = lmp.numpy
    out = get_globaldata(lmp, global_variables)
    varDict = {}
    for i, pf in enumerate(global_variables):
        varDict[pf] = out[i]
    return varDict

def get_profiledata(lmp, profile_variables, Nbins):
    "A function for extracting profile fix information"
    return get_fix_data(lmp, "Profile_variables", profile_variables, Nbins)

def get_profileDict(lmp, profile_variables, Nbins):
    """
        Otherwise we could return a dictonary 
        of named output variables as specified by the user
    """
    nlmp = lmp.numpy
    out = get_profiledata(lmp, profile_variables, Nbins)
    varDict = {}
    varDict['binNo'] = out[:, 0]
    varDict['Ncount'] = out[:, 1]
    for i, pf in enumerate(profile_variables):
        varDict[pf] = out[:, i]
    return varDict

def save_state(lmp, statename, save_variables=["x", "y", "z", "vx", "vy", "vz"]):

    state = {}
    state['name'] = statename
    state['save_variables'] = save_variables
    cmdstr = "fix " + statename + " all store/state 0 {}".format(' '.join(save_variables))
    lmp.command(cmdstr)

    return state

def load_state(lmp, state):
    
    cmdstr = "change_box all  xy final 0\n"
    for i, s in enumerate(state['save_variables']):
        varname = "p"+s 
        cmdstr += "variable " + varname + " atom f_"+state['name']+"["+str(i+1)+"]\n"
        cmdstr += "set             atom * " + s + " v_"+varname+"\n"

    for line in cmdstr.split("\n"):
        lmp.command(line)
    return None


def set_list(lmp, setlist):
    
    for s in setlist:
        lmp.command(s)

def unset_list(lmp, setlist):

    for s in setlist:
        w = s.split()
        if w[0] == "compute":
            lmp.command("uncompute " + w[1])
        elif w[0] == "fix":
            lmp.command("unfix " + w[1])
        else:
            #Not need to unset variables
            pass

def apply_mapping(lmp, map_index):

    map_list=["x","y","z"]
    N=len(map_list)

    ind=map_index

    cmdstr=""
    for i in range(N):

        mp = ind % 2
        ind = np.floor(ind/2)

        cmdstr += "variable map atom  v"+map_list[N-1-i]+"-(2*v"+map_list[N-1-i]+"*"+str(mp)+")\n"
        cmdstr += "set atom * v"+map_list[N-1-i]+" v_map\n"


        mp=ind % 2
        ind = np.floor(ind/2)

        cmdstr += "variable map atom  "+map_list[N-1-i]+"+(("+map_list[N-1-i]+"hi-2*"+map_list[N-1-i]+")*"+str(mp)+")\n"
        cmdstr += "set atom * "+map_list[N-1-i]+" v_map\n"

    for line in cmdstr.split("\n"):
        lmp.command(line)

    return None

def sum_prev_dt(A, t):
    return (   A[t-2,...] 
            +4*A[t-1,...] 
            +  A[t  ,...])/3.

def TTCF_integration(A, int_step):
 
    integral = np.zeros(A.shape)
    N = A.shape[0]

    for t in range(2 , N, 2):
        integral[t,...]   = integral[t-2,...] +  int_step*sum_prev_dt(A, t)
        integral[t-1,...] = (integral[t-2,...] + integral[t,...])/2

    if (N % 2) == 0:
        integral[-1,...] = integral[-2,...] + int_step/2*A[-1,...]

    return integral






import unittest

class TestLAMMPSImport(unittest.TestCase):
    def test_import_lammps(self):
        try:
            import lammps
            self.assertTrue(True, "PyLAMMPS import successful")
        except ImportError as e:
            self.fail(f"Failed to import PyLAMMPS: {e}")

if __name__ == '__main__':
    unittest.main()



import numpy as np

from .utils import *

class TTCF():

    def __init__(self, global_variables, profile_variables, 
                 Nsteps, Nbins, Nmappings):

        """
            Create a TTCF object to collect data

            inputs include
            global_variables - list - variables collected by ave/time
            profile_variables - list - variables collected by ave/chunk
            Nsteps - int - Number of steps daughter us run for when TTCF is collected
            Nbins - int - Number of bins in profiles
            Nmappings - int - Number of mappings used by duaghters
        """

        self.global_variables = global_variables
        self.profile_variables = profile_variables
        self.Nsteps = Nsteps
        self.Nbins = Nbins
        self.Nmappings = Nmappings
        self.Count = 0

        avetime_ncol = len(global_variables)
        avechunk_ncol = len(profile_variables) + 2

        #Allocate arrays to store data
        self.DAV_global_mean  = np.zeros([Nsteps, avetime_ncol])
        self.DAV_profile_mean   = np.zeros([Nsteps, Nbins, avechunk_ncol])

        self.TTCF_global_mean  = np.zeros([Nsteps, avetime_ncol])
        self.TTCF_profile_mean   = np.zeros([Nsteps, Nbins, avechunk_ncol])

        self.DAV_global_var  = np.zeros([Nsteps, avetime_ncol])
        self.DAV_profile_var   = np.zeros([Nsteps, Nbins, avechunk_ncol])

        self.TTCF_global_var  = np.zeros([Nsteps, avetime_ncol])
        self.TTCF_profile_var   = np.zeros([Nsteps, Nbins, avechunk_ncol])

        self.TTCF_global_partial = np.zeros([Nsteps, avetime_ncol])
        self.TTCF_profile_partial= np.zeros([Nsteps, Nbins, avechunk_ncol])

        self.DAV_profile_partial= np.zeros([Nsteps, Nbins, avechunk_ncol])
        self.DAV_global_partial = np.zeros([Nsteps, avetime_ncol])

        self.integrand_global_partial  = np.zeros([Nsteps, avetime_ncol])
        self.integrand_profile_partial = np.zeros([Nsteps, Nbins, avechunk_ncol])

    def add_mappings(self, data_profile, data_global, omega):

        #Sum the mappings together
        self.DAV_profile_partial  += data_profile[:,:,:]
        self.DAV_global_partial   += data_global[:,:]
        
        self.integrand_profile_partial += data_profile[:,:,:]*omega
        self.integrand_global_partial  += data_global[:,:]*omega

    def integrate(self, step):

        #Perform the integration
        self.TTCF_profile_partial = TTCF_integration(self.integrand_profile_partial, step)
        self.TTCF_global_partial  = TTCF_integration(self.integrand_global_partial, step)

        #Add the initial value (t=0) 
        self.TTCF_profile_partial += self.DAV_profile_partial[0,:,:]
        self.TTCF_global_partial  += self.DAV_global_partial[0,:]

        #Average over the mappings and update the Count (# of children trajectories generated excluding the mappings)
        self.DAV_profile_partial  /= self.Nmappings   
        self.DAV_global_partial   /= self.Nmappings 
        self.TTCF_profile_partial /= self.Nmappings   
        self.TTCF_global_partial  /= self.Nmappings 

        self.Count += 1

        #Update all means and variances
        if self.Count >1:
        
            self.TTCF_profile_var= update_var(self.TTCF_profile_partial, self.TTCF_profile_mean, self.TTCF_profile_var, self.Count)      
            self.DAV_profile_var= update_var(self.DAV_profile_partial, self.DAV_profile_mean, self.DAV_profile_var, self.Count)
            self.TTCF_global_var= update_var(self.TTCF_global_partial, self.TTCF_global_mean, self.TTCF_global_var, self.Count)   
            self.DAV_global_var= update_var(self.DAV_global_partial, self.DAV_global_mean, self.DAV_global_var, self.Count)
          
        self.TTCF_profile_mean= update_mean(self.TTCF_profile_partial, self.TTCF_profile_mean, self.Count)     
        self.DAV_profile_mean= update_mean(self.DAV_profile_partial, self.DAV_profile_mean, self.Count)
        self.TTCF_global_mean= update_mean(self.TTCF_global_partial, self.TTCF_global_mean, self.Count)
        self.DAV_global_mean= update_mean(self.DAV_global_partial, self.DAV_global_mean, self.Count)

        self.DAV_profile_partial[:,:,:] = 0
        self.DAV_global_partial[:,:]    = 0
            
        self.integrand_profile_partial[:,:,:] = 0
        self.integrand_global_partial[:,:]    = 0

    def finalise_output(self, irank, comm, root=0):

        self.irank = irank
        self.comm = comm
        self.root = root
        self.nprocs = comm.Get_size()
        self.output_finalised = True

        #Get FINAL COLUMN BECAUSE BY DEFAULT LAMMPS GIVE YOU ALSO THE USELESS INFO ABOUT THE BINS.
        # For  N QUANTITIES, could TAKE THE LAST N ELEMENTS
        self.TTCF_profile_mean = self.TTCF_profile_mean[:,:,2:]
        self.DAV_profile_mean  = self.DAV_profile_mean[:,:,2:]

        self.TTCF_profile_var = self.TTCF_profile_var[:,:,2:]
        self.DAV_profile_var  = self.DAV_profile_var[:,:,2:]

        self.TTCF_global_var/= float(self.Count)
        self.DAV_global_var /= float(self.Count)
        self.TTCF_profile_var /= float(self.Count)
        self.DAV_profile_var  /= float(self.Count)

        #Compute MEN AND VARIANCE OF BOTH DAV AND TTCF
        self.TTCF_profile_mean_total = sum_over_MPI(self.TTCF_profile_mean, irank, comm)
        self.DAV_profile_mean_total = sum_over_MPI(self.DAV_profile_mean, irank, comm)
        self.TTCF_profile_var_total = sum_over_MPI(self.TTCF_profile_var, irank, comm)
        self.DAV_profile_var_total = sum_over_MPI(self.DAV_profile_var, irank, comm)

        self.TTCF_global_mean_total = sum_over_MPI(self.TTCF_global_mean, irank, comm)
        self.DAV_global_mean_total = sum_over_MPI(self.DAV_global_mean, irank, comm)
        self.TTCF_global_var_total = sum_over_MPI(self.TTCF_global_var, irank, comm)
        self.DAV_global_var_total = sum_over_MPI(self.DAV_global_var, irank, comm)

        #Total is None on everything but the root processor
        if irank == self.root:
            self.TTCF_profile_mean_total = self.TTCF_profile_mean_total/float(self.nprocs)
            self.DAV_profile_mean_total  = self.DAV_profile_mean_total/float(self.nprocs)
            self.TTCF_profile_var_total  = self.TTCF_profile_var_total/float(self.nprocs)
            self.DAV_profile_var_total   = self.DAV_profile_var_total/float(self.nprocs)
            
            self.TTCF_global_mean_total = self.TTCF_global_mean_total/float(self.nprocs)
            self.DAV_global_mean_total  = self.DAV_global_mean_total/float(self.nprocs)
            self.TTCF_global_var_total  = self.TTCF_global_var_total/float(self.nprocs)
            self.DAV_global_var_total   = self.DAV_global_var_total/float(self.nprocs)
            
            self.TTCF_profile_SE_total  = np.sqrt(self.TTCF_profile_var_total)
            self.DAV_profile_SE_total   = np.sqrt(self.DAV_profile_var_total)
            self.TTCF_global_SE_total   = np.sqrt(self.TTCF_global_var_total)
            self.DAV_global_SE_total    = np.sqrt(self.DAV_global_var_total)


    def plot_data(self, animated=False):

        if self.output_finalised and self.irank == self.root:

            import matplotlib.pyplot as plt

            if animated:
                # This code animates the time history
                plt.ion()
                fig, ax = plt.subplots(1,1)
                plt.show()
                ft = True
                for t in range(self.TTCF_profile_mean_total.shape[0]):
                    print(t)
                    l1, = ax.plot(self.DAV_profile_mean_total[t, :],'r-', label="DAV")
                    l2, = ax.plot(self.TTCF_profile_mean_total[t, :],'b-', label="TTCF")
                    if ft:
                        plt.legend()
                        ft=False
                    plt.pause(0.1)
                    l1.remove()
                    l2.remove()
            else:
                #This code plots the average over time
                plt.plot(np.mean(self.DAV_profile_mean_total[:, :],0),'r-', label="DAV")
                plt.plot(np.mean(self.TTCF_profile_mean_total[:, :],0),'b-', label="TTCF")
                plt.legend()
                plt.show()

    def save_data(self):

        if self.output_finalised and self.irank == self.root:
            #print a separate file for each profile variable
            for i in range(len(self.profile_variables)):
            
                var_name=self.profile_variables[i]
                #replace "/" character in the name to avoid crashes (it is read as a folder)
                var_name=var_name.replace('/', '_')
             
                print(var_name)
                
                np.savetxt('profile_DAV_' + var_name + '.txt', self.DAV_profile_mean_total[:,:,i])
                np.savetxt('profile_TTCF_' + var_name + '.txt', self.TTCF_profile_mean_total[:,:,i])
                np.savetxt('profile_DAV_SE_' + var_name + '.txt', self.DAV_profile_SE_total[:,:,i])
                np.savetxt('profile_TTCF_SE_' + var_name + '.txt',self.TTCF_profile_SE_total[:,:,i])
            
            
            #np.savetxt('profile_DAV.txt', self.DAV_profile_mean_total)
            #np.savetxt('profile_TTCF.txt', self.TTCF_profile_mean_total)
            
            #np.savetxt('profile_DAV_SE.txt', self.DAV_profile_SE_total)
            #np.savetxt('profile_TTCF_SE.txt', self.TTCF_profile_SE_total)
            
            np.savetxt('global_DAV.txt', self.DAV_global_mean_total)
            np.savetxt('global_TTCF.txt', self.TTCF_global_mean_total)
            
            np.savetxt('global_DAV_SE.txt', self.DAV_global_SE_total)
            np.savetxt('global_TTCF_SE.txt', self.TTCF_global_SE_total)
                 




#!/usr/bin/env python

from setuptools import setup, find_packages

# To use a consistent encoding
from codecs import open
from os import path

from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup( name = "TTCF4LAMMPS",
       version = "1.0.0",
       author = ["Luca Maffioli","Edward Smith"],
       author_email = "lmaffioli@swin.edu.au",
       url = "https://github.com/edwardsmith999/TTCF4LAMMPS",
       classifiers=['Development Status :: 3 - Alpha',
                     'Programming Language :: Python :: 3.11'],
       packages=find_packages(exclude=['contrib', 'docs', 'tests']),
       keywords='Molecular Dynamics TTCF Transient Time Correlation Function',
       license = "GPL",
       install_requires=['numpy', 'scipy', 'matplotlib', 'mpi4py', 'lammps' ],
       description = "Code to run TTCF function on LAMMPS",
       long_description = long_description,
       long_description_content_type='text/markdown',
       entry_points={
            'console_scripts': [
                'TTCF=run_TTCF:main',
            ],
       },
)
